/** @file
 *
 *  A brief file description
 *
 *  @section license License
 *
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#include "catch.hpp"

#include "proxy/http3/Http3FrameHandler.h"

class Http3MockFrameHandler : public Http3FrameHandler
{
public:
  int total_frame_received = 0;

  // Http3FrameHandler

  std::vector<Http3FrameType>
  interests() override
  {
    return {Http3FrameType::DATA, Http3FrameType::SETTINGS};
  }

  Http3ErrorUPtr
  handle_frame(std::shared_ptr<const Http3Frame> /* frame ATS_UNUSED */, Http3StreamType /* s_type */) override
  {
    this->total_frame_received++;
    return Http3ErrorUPtr(nullptr);
  }
};
