/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#pragma once

#include <string.h>

#include <swoc/swoc_ip.h>
#include <yaml-cpp/yaml.h>

#include "ts/ts.h"

// ToDo: Maybe in the future this should be a template class, but for now it's a simple
// subclass wrapper over swoc::IPRangeSet.
namespace List
{
class IP : public swoc::IPRangeSet
{
  using self_type = IP;

public:
  explicit IP(std::string &name) : _name(name) {}

  IP()                                    = delete;
  IP(self_type &&)                        = delete;
  self_type &operator=(const self_type &) = delete;
  self_type &operator=(self_type &&)      = delete;

  void
  add(std::string &str)
  {
    if (swoc::IPRange r; r.load(str)) {
      mark(r);
    } else {
      TSReleaseAssert("Bad IP range");
    }
  }

  bool parseYaml(const YAML::Node &node);

  const std::string &
  name() const
  {
    return _name;
  }

private:
  std::string _name;

}; // class IpList

} // namespace List
