#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

sed 's/ ___FS___ /\
/g' | sed 's/^{//' | sed 's/}$//' | while read LN
do
    # For lines that contain a full set of message headers, get
    # individual headers on separate lines, then sort the lines.
    #
    echo $LN | sed 's/}}/}}\
/g' | sort --ignore-nonprinting  --dictionary-order
done
