/*
  Licensed to the Apache Software Foundation (ASF) under one
  or more contributor license agreements.  See the NOTICE file
  distributed with this work for additional information
  regarding copyright ownership.  The ASF licenses this file
  to you under the Apache License, Version 2.0 (the
  "License"); you may not use this file except in compliance
  with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
*/

#pragma once

#include "ts_lua_common.h"

struct fetch_multi_info;

typedef struct {
  TSCont                   contp;
  struct fetch_multi_info *fmi;

  TSIOBuffer       buffer;
  TSIOBufferReader reader;
  TSFetchSM        fch;

  unsigned int over   : 1;
  unsigned int failed : 1;
} ts_lua_fetch_info;

typedef struct fetch_multi_info {
  TSCont            contp; // should be destroyed only in cleanup
  int               multi; // invoke from ts.fetch_multi
  int               total;
  int               done;
  ts_lua_fetch_info fiv[0];
} ts_lua_fetch_multi_info;

void ts_lua_inject_fetch_api(lua_State *L);
