#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

N=60
while (( N > 0 ))
do
    rm -f metrics.out
    traffic_ctl metric match \.remap_stats\. \
      | grep status \
      | sort \
        > metrics.out
    sleep 1
    if diff metrics.out "${AUTEST_TEST_DIR}/gold/metrics_post.gold"
    then
        exit 0
    fi
    let N=N-1
done
echo TIMEOUT
exit 1
