/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#pragma once

#include <string>
#include <chrono>

#include "ts/ts.h"

constexpr char const PLUGIN_NAME[] = "rate_limit";

void        delayHeader(TSHttpTxn txnp, const std::string &header, std::chrono::milliseconds delay);
void        retryAfter(TSHttpTxn txnp, unsigned retry);
std::string getDescriptionFromUrl(const char *url);

namespace rate_limit_ns
{
extern DbgCtl dbg_ctl;
}
using namespace rate_limit_ns;
