/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sensitive_fields.h"

#include <cstring>
#include <catch.hpp>

using namespace traffic_dump;

TEST_CASE("SensitiveFields", "[sensitive_fields_t]")
{
  sensitive_fields_t test_fields = {"one", "tWO", "THREE"};
  CHECK(test_fields.count("one") == 1);
  CHECK(test_fields.count("oNe") == 1);

  CHECK(test_fields.count("tWO") == 1);
  CHECK(test_fields.count("two") == 1);

  CHECK(test_fields.count("THREE") == 1);
  CHECK(test_fields.count("three") == 1);
}
