/** @file

 Cache Control Extensions Cache Utilities

  @section license License

  Licensed to the Apache Software Foundation (ASF) under one
  or more contributor license agreements.  See the NOTICE file
  distributed with this work for additional information
  regarding copyright ownership.  The ASF licenses this file
  to you under the Apache License, Version 2.0 (the
  "License"); you may not use this file except in compliance
  with the License.  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */

#pragma once

#include "ts/ts.h"

struct StateInfo;
struct BodyData;

bool has_trailing_parameter(TSMBuffer buf, TSMLoc http_hdr_loc);
void add_trailing_parameter(TSMBuffer buf, TSMLoc http_hdr_loc);
bool strip_trailing_parameter(TSMBuffer buf, TSMLoc http_hdr_loc);

void      get_pristine_url(StateInfo *state);
void      fix_connection_close(StateInfo *state);
BodyData *intercept_check_request(StateInfo *state);
bool      intercept_fetch_the_url(StateInfo *state);

/*-----------------------------------------------------------------------------------------------*/
